/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_main.c

#include "quakedef.h"


r_defaultquad_t *r_default_quads = (r_defaultquad_t *) scratchbuf;
r_defaultquad2_t *r_default_quads2 = (r_defaultquad2_t *) scratchbuf;

int r_max_quads = SCRATCHBUF_SIZE / (sizeof (r_defaultquad_t) * 4);
int r_num_quads = 0;

unsigned short *r_quad_indexes = NULL;

unsigned int r_quadindexbuffer = 0;
unsigned int r_parttexcoordbuffer = 0;

void R_InitQuads (void);
void R_DrawCoronas (void);

glmatrix r_world_matrix;

void R_InitVertexBuffers (void)
{
	// create the initial indexes if needed
	if (!r_quad_indexes) R_InitQuads ();

	if (gl_arb_vertex_buffer_object)
	{
		int i;
		float *partsts = (float *) scratchbuf;
		float *st = (float *) scratchbuf;

		if (r_quadindexbuffer)
		{
			qglDeleteBuffers (1, &r_quadindexbuffer);
			r_quadindexbuffer = 0;
		}

		// generate our static index buffer for use by our Quads
		qglGenBuffers (1, &r_quadindexbuffer);
		qglBindBuffer (GL_ELEMENT_ARRAY_BUFFER, r_quadindexbuffer);
		qglBufferData (GL_ELEMENT_ARRAY_BUFFER, r_max_quads * sizeof (unsigned short) * 6, r_quad_indexes, GL_STATIC_DRAW);
		qglBindBuffer (GL_ELEMENT_ARRAY_BUFFER, 0);

		if (r_parttexcoordbuffer)
		{
			qglDeleteBuffers (1, &r_parttexcoordbuffer);
			r_parttexcoordbuffer = 0;
		}

		for (i = 0; i < r_max_quads; i++, st += 8)
		{
			st[0] = 0; st[1] = 0;
			st[2] = 1; st[3] = 0;
			st[4] = 1; st[5] = 1;
			st[6] = 0; st[7] = 1;
		}

		// generate our static vertex buffer for use by our particles and sprites
		qglGenBuffers (1, &r_parttexcoordbuffer);
		qglBindBuffer (GL_ARRAY_BUFFER, r_parttexcoordbuffer);
		qglBufferData (GL_ARRAY_BUFFER, r_max_quads * sizeof (float) * 8, partsts, GL_STATIC_DRAW);
		qglBindBuffer (GL_ARRAY_BUFFER, 0);
	}
}


void R_InitQuads (void)
{
	int i;
	unsigned short *ndx;
	unsigned int pquad;

	// don't create it more than once
	if (r_quad_indexes) return;

	// clamp to shorts
	if (r_max_quads >= 16384) r_max_quads = 16384;

	r_quad_indexes = (unsigned short *) Hunk_Alloc (r_max_quads * sizeof (unsigned short) * 6);
	ndx = r_quad_indexes;

	// set up the indexes one time only
	for (i = 0, pquad = 0; i < r_max_quads; i++, ndx += 6, pquad += 4)
	{
		ndx[0] = pquad + 0;
		ndx[1] = pquad + 1;
		ndx[2] = pquad + 2;

		ndx[3] = pquad + 0;
		ndx[4] = pquad + 2;
		ndx[5] = pquad + 3;
	}
}


void R_DrawAliasModels (void);
void R_DrawSpriteModels (void);

void R_ShowTrisBegin (void)
{
	GL_TexEnv (GL_TEXTURE0, GL_TEXTURE_2D, GL_NONE);
	glColor4f (1, 1, 1, 1);
	Fog_DisableGFog ();

	if (r_showtris.value < 2)
	{
		glDisable (GL_DEPTH_TEST);
		glDepthMask (GL_FALSE);
	}
	else
	{
		glEnable (GL_POLYGON_OFFSET_FILL);
		glEnable (GL_POLYGON_OFFSET_LINE);
		glPolygonOffset (-1, -3);
	}

	glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
	glColor4f (1, 1, 1, 1);
}


void R_ShowTrisEnd (void)
{
	glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);

	if (r_showtris.value < 2)
	{
		glDepthMask (GL_TRUE);
		glEnable (GL_DEPTH_TEST);
	}
	else
	{
		glDisable (GL_POLYGON_OFFSET_FILL);
		glDisable (GL_POLYGON_OFFSET_LINE);
	}

	GL_TexEnv (GL_TEXTURE0, GL_TEXTURE_2D, GL_REPLACE);
	Fog_EnableGFog ();
}


qboolean	r_cache_thrash;		// compatability

vec3_t		modelorg, r_entorigin;

int			r_visframecount;	// bumped when going to a new PVS
int			r_framecount;		// used for dlight push checking

glmatrix r_projection_matrix;
mplane_t	frustum[4];

// johnfitz -- rendering statistics
int rs_brushpolys, rs_aliaspolys;
int rs_drawelements;
int rs_dynamiclightmaps;

qboolean	envmap;				// true during envmap command capture

//
// view origin
//
vec3_t	vup;
vec3_t	vpn;
vec3_t	vright;
vec3_t	r_origin;

float r_fovx, r_fovy; // johnfitz -- rendering fov may be different becuase of r_waterwarp

//
// screen size info
//
refdef_t	r_refdef;

mleaf_t		*r_viewleaf, *r_oldviewleaf;

int		d_lightstylevalue[256];	// 8.8 fraction of base light value


cvar_t	r_norefresh = {"r_norefresh", "0"};
cvar_t	r_drawentities = {"r_drawentities", "1"};
cvar_t	r_drawviewmodel = {"r_drawviewmodel", "1"};
cvar_t	r_speeds = {"r_speeds", "0"};
cvar_t	r_fullbright = {"r_fullbright", "0"};
cvar_t	r_lightmap = {"r_lightmap", "0"};
cvar_t	r_shadows = {"r_shadows", "0", true};
cvar_t	r_wateralpha = {"r_wateralpha", "1", true};
cvar_t	r_dynamic = {"r_dynamic", "1", true};
cvar_t	r_novis = {"r_novis", "0", true};

cvar_t	gl_finish = {"gl_finish", "0"};
cvar_t	gl_clear = {"gl_clear", "0"};
cvar_t	gl_cull = {"gl_cull", "1"};
cvar_t	gl_smoothmodels = {"gl_smoothmodels", "1"};
cvar_t	gl_affinemodels = {"gl_affinemodels", "0"};
cvar_t	gl_polyblend = {"gl_polyblend", "1"};
cvar_t	gl_playermip = {"gl_playermip", "0"};
cvar_t	gl_nocolors = {"gl_nocolors", "0"};

// johnfitz -- new cvars
cvar_t	r_clearcolor = {"r_clearcolor", "2", true};
cvar_t	r_drawflat = {"r_drawflat", "0"};
cvar_t	r_flatlightstyles = {"r_flatlightstyles", "0"};
cvar_t	gl_fullbrights = {"gl_fullbrights", "1", true};
cvar_t	gl_overbright = {"gl_overbright", "1", true};
cvar_t	gl_overbright_models = {"gl_overbright_models", "1", true};
cvar_t	r_oldskyleaf = {"r_oldskyleaf", "0"};
cvar_t	r_drawworld = {"r_drawworld", "1"};
cvar_t	r_showbboxes = {"r_showbboxes", "0"};
cvar_t	r_lerpmodels = {"r_lerpmodels", "1"};
cvar_t	r_lerpmove = {"r_lerpmove", "1"};
cvar_t	r_nolerp_list = {"r_nolerp_list", "progs/flame.mdl,progs/flame2.mdl,progs/braztall.mdl,progs/brazshrt.mdl,progs/longtrch.mdl,progs/flame_pyre.mdl,progs/v_saw.mdl,progs/v_xfist.mdl,progs/h2stuff/newfire.mdl"};

extern cvar_t	r_vfog;
// johnfitz

cvar_t r_motionblur = {"r_motionblur", "1", true};

cvar_t r_lockpvs = {"r_lockpvs", "0"};
cvar_t r_lockfrustum = {"r_lockfrustum", "0"};

qboolean r_drawflat_cheatsafe, r_fullbright_cheatsafe, r_lightmap_cheatsafe, r_drawworld_cheatsafe; // johnfitz

qboolean r_rttwaterwarp = false;

void R_DrawViewModel (void);
void R_AlphaListBeginFrame (void);
void R_DrawAlphaList (void);

/*
=================
R_CullSphere

Returns true if the sphere is outside the frustum
grabbed from AlienArena 2011 source (repaying the compliment!!!)
=================
*/
qboolean R_CullSphere (const float *center, const float radius, const int clipflags)
{
	int		i;
	mplane_t *p;

	for (i = 0, p = frustum; i < 4; i++, p++)
	{
		if (!(clipflags & (1 << i)))
			continue;

		if (DotProduct (center, p->normal) - p->dist <= -radius)
			return true;
	}

	return false;
}


/*
=================
R_CullBox -- johnfitz -- replaced with new function from lordhavoc

Returns true if the box is completely outside the frustum
=================
*/
qboolean R_CullBox (vec3_t emins, vec3_t emaxs)
{
	int i;
	mplane_t *p;

	for (i = 0; i < 4; i++)
	{
		p = &frustum[i];

		switch (p->signbits)
		{
		default:
		case 0:
			if (p->normal[0] * emaxs[0] + p->normal[1] * emaxs[1] + p->normal[2] * emaxs[2] < p->dist)
				return true;

			break;

		case 1:
			if (p->normal[0] * emins[0] + p->normal[1] * emaxs[1] + p->normal[2] * emaxs[2] < p->dist)
				return true;

			break;

		case 2:
			if (p->normal[0] * emaxs[0] + p->normal[1] * emins[1] + p->normal[2] * emaxs[2] < p->dist)
				return true;

			break;

		case 3:
			if (p->normal[0] * emins[0] + p->normal[1] * emins[1] + p->normal[2] * emaxs[2] < p->dist)
				return true;

			break;

		case 4:
			if (p->normal[0] * emaxs[0] + p->normal[1] * emaxs[1] + p->normal[2] * emins[2] < p->dist)
				return true;

			break;

		case 5:
			if (p->normal[0] * emins[0] + p->normal[1] * emaxs[1] + p->normal[2] * emins[2] < p->dist)
				return true;

			break;

		case 6:
			if (p->normal[0] * emaxs[0] + p->normal[1] * emins[1] + p->normal[2] * emins[2] < p->dist)
				return true;

			break;

		case 7:
			if (p->normal[0] * emins[0] + p->normal[1] * emins[1] + p->normal[2] * emins[2] < p->dist)
				return true;

			break;
		}
	}

	return false;
}
/*
===============
R_CullModelForEntity -- johnfitz -- uses correct bounds based on rotation
===============
*/
qboolean R_CullModelForEntity (entity_t *e)
{
	vec3_t mins, maxs;

	if (e->angles[0] || e->angles[2]) //pitch or roll
	{
		VectorAdd (e->origin, e->model->rmins, mins);
		VectorAdd (e->origin, e->model->rmaxs, maxs);
	}
	else if (e->angles[1]) //yaw
	{
		VectorAdd (e->origin, e->model->ymins, mins);
		VectorAdd (e->origin, e->model->ymaxs, maxs);
	}
	else //no rotation
	{
		VectorAdd (e->origin, e->model->mins, mins);
		VectorAdd (e->origin, e->model->maxs, maxs);
	}

	if (e->scale != 16)
	{
		// scale the bbox
		VectorScale (mins, (float) e->scale / 16.0f, mins);
		VectorScale (maxs, (float) e->scale / 16.0f, maxs);
	}

	return R_CullBox (mins, maxs);
}


//==============================================================================
//
// SETUP FRAME
//
//==============================================================================

int SignbitsForPlane (mplane_t *out)
{
	int	bits, j;

	// for fast box on planeside test
	bits = 0;

	for (j = 0; j < 3; j++)
	{
		if (out->normal[j] < 0)
			bits |= 1 << j;
	}

	return bits;
}


/*
=============
GL_SetFrustum
=============
*/
void GL_SetFrustum (float fovx, float fovy)
{
#define NEARCLIP 4
	float left, right, bottom, top;
	float nudge = 1.0 - 1.0 / (1 << 23);

	right = NEARCLIP * tan (fovx * M_PI / 360.0);
	left = -right;

	top = NEARCLIP * tan (fovy * M_PI / 360.0);
	bottom = -top;

	r_projection_matrix.m16[0] = (2 * NEARCLIP) / (right - left);
	r_projection_matrix.m16[4] = 0;
	r_projection_matrix.m16[8] = (right + left) / (right - left);
	r_projection_matrix.m16[12] = 0;

	r_projection_matrix.m16[1] = 0;
	r_projection_matrix.m16[5] = (2 * NEARCLIP) / (top - bottom);
	r_projection_matrix.m16[9] = (top + bottom) / (top - bottom);
	r_projection_matrix.m16[13] = 0;

	r_projection_matrix.m16[2] = 0;
	r_projection_matrix.m16[6] = 0;
	r_projection_matrix.m16[10] = -1 * nudge;
	r_projection_matrix.m16[14] = -2 * NEARCLIP * nudge;

	r_projection_matrix.m16[3] = 0;
	r_projection_matrix.m16[7] = 0;
	r_projection_matrix.m16[11] = -1;
	r_projection_matrix.m16[15] = 0;

	if (r_waterwarp.value && r_waterwarp.value < 2)
	{
		int contents = Mod_PointInLeaf (r_origin, cl.worldmodel)->contents;

		if ((contents == CONTENTS_WATER || contents == CONTENTS_SLIME || contents == CONTENTS_LAVA || cl.inwater) && v_blend[3])
		{
			r_projection_matrix.m16[4] = sin (cl.time * 1.125f) * 0.0666f;
			r_projection_matrix.m16[1] = cos (cl.time * 1.125f) * 0.0666f;
			GL_ScaleMatrix (&r_projection_matrix, (cos (cl.time * 0.75f) + 20.0f) * 0.05f, (sin (cl.time * 0.75f) + 20.0f) * 0.05f, 1);
		}
	}

	glLoadMatrixf (r_projection_matrix.m16);
}


void R_ExtractFrustum (void)
{
	int i;
	glmatrix r_mvp_matrix;

	// get the frustum from the actual matrixes that were used
	GL_MultiplyMatrix (&r_mvp_matrix, &r_world_matrix, &r_projection_matrix);

	// frustum 0
	frustum[0].normal[0] = r_mvp_matrix.m16[3] - r_mvp_matrix.m16[0];
	frustum[0].normal[1] = r_mvp_matrix.m16[7] - r_mvp_matrix.m16[4];
	frustum[0].normal[2] = r_mvp_matrix.m16[11] - r_mvp_matrix.m16[8];

	// frustum 1
	frustum[1].normal[0] = r_mvp_matrix.m16[3] + r_mvp_matrix.m16[0];
	frustum[1].normal[1] = r_mvp_matrix.m16[7] + r_mvp_matrix.m16[4];
	frustum[1].normal[2] = r_mvp_matrix.m16[11] + r_mvp_matrix.m16[8];

	// frustum 2
	frustum[2].normal[0] = r_mvp_matrix.m16[3] + r_mvp_matrix.m16[1];
	frustum[2].normal[1] = r_mvp_matrix.m16[7] + r_mvp_matrix.m16[5];
	frustum[2].normal[2] = r_mvp_matrix.m16[11] + r_mvp_matrix.m16[9];

	// frustum 3
	frustum[3].normal[0] = r_mvp_matrix.m16[3] - r_mvp_matrix.m16[1];
	frustum[3].normal[1] = r_mvp_matrix.m16[7] - r_mvp_matrix.m16[5];
	frustum[3].normal[2] = r_mvp_matrix.m16[11] - r_mvp_matrix.m16[9];

	for (i = 0; i < 4; i++)
	{
		VectorNormalize (frustum[i].normal);

		frustum[i].type = PLANE_ANYZ;
		frustum[i].dist = DotProduct (r_origin, frustum[i].normal); //FIXME: shouldn't this always be zero?
		frustum[i].signbits = SignbitsForPlane (&frustum[i]);
	}
}


gltexture_t *gl_warp_texture;

int gl_rtt_width = 0;
int gl_rtt_height = 0;

gltexture_t *R_CreateRTTTexture (char *name, byte *data)
{
	int i;
	gltexture_t *tex = NULL;

	for (i = 0; i < gl_rtt_width * gl_rtt_height; i++)
		data[i] = rand () & 255;

	// now create the texture for it
	tex = TM_LoadImage
	(
		NULL,
		name,
		gl_rtt_width,
		gl_rtt_height,
		SRC_INDEXED,
		data,
		"",
		(src_offset_t) data,
		TEXPREF_OVERWRITE | TEXPREF_NOPICMIP | TEXPREF_PERSIST | TEXPREF_LINEAR
	);

	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	return tex;
}


void R_CreateRTTTextures (int width, int height)
{
	int maxrtt;
	static byte *rtttexels = NULL;

	// figure the size to create the texture at
	glGetIntegerv (GL_MAX_TEXTURE_SIZE, &maxrtt);

	// never create larger than this
	if (maxrtt > 1024) maxrtt = 1024;

	// scale to powers of 2
	for (gl_rtt_width = 1; gl_rtt_width < width; gl_rtt_width <<= 1);
	for (gl_rtt_height = 1; gl_rtt_height < height; gl_rtt_height <<= 1);

    // round down
    //if (gl_rtt_width > width) gl_rtt_width >>= 1;
    //if (gl_rtt_height > height) gl_rtt_height >>= 1;

	// clamp to max texture size
	if (gl_rtt_width > maxrtt) gl_rtt_width = maxrtt;
	if (gl_rtt_height > maxrtt) gl_rtt_height = maxrtt;

	if (gl_warp_texture) TM_FreeTexture (gl_warp_texture);
	if (rtttexels) free (rtttexels);

	// never freed so that the texture can reload properly in a vid_restart
	rtttexels = (byte *) malloc (gl_rtt_width * gl_rtt_height);

	// create the textures for RTT effects
	gl_warp_texture = R_CreateRTTTexture ("gl_warp_texture", rtttexels);

	Con_Printf ("Created rtt textures at %ix%i\n", gl_rtt_width, gl_rtt_height);
}


/*
=============
R_SetupGL
=============
*/
int gl_vpx, gl_vpy, gl_vpwidth, gl_vpheight;

void R_SetupGL (void)
{
	// evaluate viewport size
	gl_vpx = 0;
	gl_vpy = sb_lines;
	gl_vpwidth = glwidth;
	gl_vpheight = glheight - sb_lines;

	if (r_rttwaterwarp)
	{
		// bound viewport dimensions to the rtt texture size
		if (gl_vpwidth > gl_rtt_width) gl_vpwidth = gl_rtt_width;
		if (gl_vpheight > gl_rtt_height) gl_vpheight = gl_rtt_height;
	}
}


void R_SetViewport (void)
{
	// now set the correct viewport
	glViewport (gl_vpx, gl_vpy, gl_vpwidth, gl_vpheight);

	// johnfitz -- rewrote this section
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	// johnfitz

	GL_SetFrustum (r_fovx, r_fovy); // johnfitz -- use r_fov* vars

	glMatrixMode (GL_MODELVIEW);

	GL_IdentityMatrix (&r_world_matrix);

	GL_RotateMatrix (&r_world_matrix, -90, 1, 0, 0);
	GL_RotateMatrix (&r_world_matrix, 90, 0, 0, 1);

	GL_RotateMatrix (&r_world_matrix, -r_refdef.viewangles[2], 1, 0, 0);
	GL_RotateMatrix (&r_world_matrix, -r_refdef.viewangles[0], 0, 1, 0);
	GL_RotateMatrix (&r_world_matrix, -r_refdef.viewangles[1], 0, 0, 1);

	GL_TranslateMatrix (&r_world_matrix, -r_refdef.vieworg[0], -r_refdef.vieworg[1], -r_refdef.vieworg[2]);

	glLoadMatrixf (r_world_matrix.m16);

	R_ExtractFrustum ();

	// set drawing parms
	if (gl_cull.value)
		glEnable (GL_CULL_FACE);
	else glDisable (GL_CULL_FACE);

	glDisable (GL_BLEND);
	glDisable (GL_ALPHA_TEST);
	glEnable (GL_DEPTH_TEST);
	glDepthMask (GL_TRUE);
	glDepthFunc (GL_LEQUAL);
}


/*
=============
R_Clear -- johnfitz -- rewritten and gutted
=============
*/
void R_SetClearColor_f (void);

void R_Clear (void)
{
	extern int gl_stencilbits;
	unsigned int clearbits;

	// partition the depth range for 3D/2D stuff
	glDepthRange (QGL_DEPTH_3D_BEGIN, QGL_DEPTH_3D_END);

	// always clear the depth buffer
	clearbits = GL_DEPTH_BUFFER_BIT;

	if (r_viewleaf && r_viewleaf->contents == CONTENTS_SOLID)
	{
		// when noclipping set a black clear color
		glClearColor (0.15, 0.15, 0.15, 1);
		clearbits |= GL_COLOR_BUFFER_BIT;
	}
	else if (gl_clear.value)
	{
		// set the correct clear color
		R_SetClearColor_f ();
		clearbits |= GL_COLOR_BUFFER_BIT;
	}

	// if we got a stencil buffer we must always clear it, even if we're not using it
	if (gl_stencilbits) clearbits |= GL_STENCIL_BUFFER_BIT;

	glClear (clearbits);
}

/*
===============
R_SetupScene -- johnfitz -- this is the stuff that needs to be done once per eye in stereo mode
===============
*/
void R_SetupScene (void)
{
	R_AnimateLight ();
	r_framecount++;
}


void R_UpdateContentsColor (void)
{
	/*
	if (r_viewleaf->contents == CONTENTS_EMPTY)
	{
		// brought to you today by "disgusting hacks 'r' us"...
		extern cshift_t cshift_empty;

		cl.cshifts[CSHIFT_CONTENTS].destcolor[0] = cshift_empty.destcolor[0];
		cl.cshifts[CSHIFT_CONTENTS].destcolor[1] = cshift_empty.destcolor[1];
		cl.cshifts[CSHIFT_CONTENTS].destcolor[2] = cshift_empty.destcolor[2];
		cl.cshifts[CSHIFT_CONTENTS].percent = cshift_empty.percent;
	}
	else if (r_viewleaf->contentscolor)
	{
		// if the viewleaf has a contents colour set we just override with it
		cl.cshifts[CSHIFT_CONTENTS].destcolor[0] = r_viewleaf->contentscolor[0];
		cl.cshifts[CSHIFT_CONTENTS].destcolor[1] = r_viewleaf->contentscolor[1];
		cl.cshifts[CSHIFT_CONTENTS].destcolor[2] = r_viewleaf->contentscolor[2];

		// these now have more colour so reduce the percent to compensate
		if (r_viewleaf->contents == CONTENTS_WATER)
			cl.cshifts[CSHIFT_CONTENTS].percent = 48;
		else if (r_viewleaf->contents == CONTENTS_LAVA)
			cl.cshifts[CSHIFT_CONTENTS].percent = 112;
		else if (r_viewleaf->contents == CONTENTS_SLIME)
			cl.cshifts[CSHIFT_CONTENTS].percent = 80;
		else cl.cshifts[CSHIFT_CONTENTS].percent = 0;
	}
	else
	*/
	{
		// empty or undefined
		V_SetContentsColor (r_viewleaf->contents);
	}

	// now calc the blend
	V_CalcBlend ();
}


/*
===============
R_SetupView -- johnfitz -- this is the stuff that needs to be done once per frame, even in stereo mode
===============
*/
void R_SetupView (void)
{
	Fog_SetupFrame (); // johnfitz

	// build the transformation matrix for the given view angles
	VectorCopy (r_refdef.vieworg, r_origin);
	AngleVectors (r_refdef.viewangles, vpn, vright, vup);

	// current viewleaf
	r_oldviewleaf = r_viewleaf;
	r_viewleaf = Mod_PointInLeaf (r_origin, cl.worldmodel);

	// mh - replaced with new version
	R_UpdateContentsColor ();
	//V_SetContentsColor (r_viewleaf->contents);
	//V_CalcBlend ();

	r_cache_thrash = false;

	// johnfitz -- calculate r_fovx and r_fovy here
	r_fovx = r_refdef.fov_x;
	r_fovy = r_refdef.fov_y;

	if (r_waterwarp.value)
	{
		int contents = Mod_PointInLeaf (r_origin, cl.worldmodel)->contents;

		if ((contents == CONTENTS_WATER || contents == CONTENTS_SLIME || contents == CONTENTS_LAVA || cl.inwater) && v_blend[3])
		{
			if (r_waterwarp.value > 2)
			{
				// variance is a percentage of width, where width = 2 * tan(fov / 2) otherwise the effect is too dramatic at high FOV and too subtle at low FOV.  what a mess!
				r_fovx = atan (tan (DEG2RAD (r_refdef.fov_x) / 2) * (0.97 + sin (cl.time * 1.5) * 0.03)) * 2 / M_PI_DIV_180;
				r_fovy = atan (tan (DEG2RAD (r_refdef.fov_y) / 2) * (1.03 - sin (cl.time * 1.5) * 0.03)) * 2 / M_PI_DIV_180;
			}
			else if (r_waterwarp.value > 1) r_rttwaterwarp = true;
		}
	}

	// johnfitz

	R_MarkLeafs (); // johnfitz -- create texture chains from PVS

	// johnfitz -- cheat-protect some draw modes
	// (isn't client-side cheat-protection kinda bogus?)  the cheater can just remove it and recompile, after all.
	r_drawflat_cheatsafe = r_fullbright_cheatsafe = r_lightmap_cheatsafe = false;
	r_drawworld_cheatsafe = true;

	if (cl.maxclients == 1)
	{
		if (!r_drawworld.value) r_drawworld_cheatsafe = false;

		if (r_drawflat.value) r_drawflat_cheatsafe = true;
		else if (r_fullbright.value || !cl.worldmodel->lightdata) r_fullbright_cheatsafe = true;
		else if (r_lightmap.value) r_lightmap_cheatsafe = true;
	}

	// johnfitz
}

//==============================================================================
//
// RENDER VIEW
//
//==============================================================================

/*
================
R_RenderScene
================
*/
// mh 2010-07-16
void R_ModelSurfsBeginFrame (void);

void RTTWarp_Setup (float *warpblend)
{
	GL_BindTexture (GL_TEXTURE0, gl_warp_texture);
	glCopyTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, 0, sb_lines, gl_vpwidth, gl_vpheight);

	// rescale texcoords to viewport/texture dimensions
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glScalef ((float) gl_vpwidth / (float) gl_rtt_width, (float) gl_vpheight / (float) gl_rtt_height, 1);
	glMatrixMode (GL_MODELVIEW);

	// ensure
	if (gl_polyblend.value && 0)
	{
		/*
		float lerpblend[4] = {1, 1, 1, 1};

		// set blend for interpolation
		lerpblend[3] = 1.0f - warpblend[3];

		// this is the proper blend mode for underwater colours
		glColor4f (warpblend[0], warpblend[1], warpblend[2], 1);
		glTexEnvfv (GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, lerpblend);
		GL_TexEnv (GL_TEXTURE0, GL_TEXTURE_2D, GL_INTERPOLATE);
		*/
		glColor4f (warpblend[0], warpblend[1], warpblend[2], warpblend[3]);
		GL_TexEnv (GL_TEXTURE0, GL_TEXTURE_2D, GL_BLEND);
	}
	else GL_TexEnv (GL_TEXTURE0, GL_TEXTURE_2D, GL_REPLACE);
}


void RTTWarp_Takedown (void)
{
	if (gl_polyblend.value && 0)
	{
		float warpblend[4] = {1, 1, 1, 1};

		glTexEnvfv (GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, warpblend);
		GL_TexEnv (GL_TEXTURE0, GL_TEXTURE_2D, GL_REPLACE);
		glColor4f (1, 1, 1, 1);
	}

	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);
}


#define WARP_TESS	16

void R_DrawUnderwaterWarp (void)
{
	int numindexes = WARP_TESS * WARP_TESS * 6;
	int x, y, i, ndx;
	float *dst = NULL;

	float hscale, invhscale;
	float textessw, textessh;
	float invtess, scale;
	float rdt;
	unsigned short *indexes = (unsigned short *) scratchbuf;
	float *vertexes = (float *) (indexes + 32768);

	// no underwater warp
	if (!r_rttwaterwarp) return;

	// for the next frame - done first so it will be validly false if we return from the next 2
	r_rttwaterwarp = false;

	if (!r_viewleaf) return;
	if (cls.state != ca_connected) return;

	RTTWarp_Setup (v_blend);

	hscale = (float) (glheight - sb_lines) / (float) glheight;
	invhscale = 1.0f / hscale;
	textessw = (float) WARP_TESS;
	textessh = (float) WARP_TESS * hscale;
	invtess = 1.0f / WARP_TESS;
	rdt = cl.time * 2.0f;
	scale = (0.125f * (float) WARP_TESS) / 32.0f;

	dst = vertexes;

	GL_SetStreamSource (0, GLSTREAM_POSITION, 3, GL_FLOAT, 5 * sizeof (float), dst);
	GL_SetStreamSource (0, GLSTREAM_COLOR, 0, GL_NONE, 0, NULL);
	GL_SetStreamSource (0, GLSTREAM_TEXCOORD0, 2, GL_FLOAT, 5 * sizeof (float), &dst[3]);
	GL_SetStreamSource (0, GLSTREAM_TEXCOORD1, 0, GL_NONE, 0, NULL);
	GL_SetStreamSource (0, GLSTREAM_TEXCOORD2, 0, GL_NONE, 0, NULL);

	// fixme - could we avoid a lot of this crap if we just reused the vertexes as texcoords, warped them instead, and
	// scaled them in the texture matrix?  (likewise scale the texture matrix to flip the texture)
	for (y = 0; y <= WARP_TESS; y++)
	{
		for (x = 0; x <= WARP_TESS; x++, dst += 5)
		{
			float s = invtess * x * textessw;
			float t = (invtess * y) * textessh;

			dst[0] = (float) (x * glwidth / WARP_TESS);
			dst[1] = (float) ((y * glheight / WARP_TESS) * hscale);
			dst[2] = 0;

			if (x == 0 || x == WARP_TESS)
				dst[3] = (float) dst[0] / (float) glwidth;
			else dst[3] = (s + sin (t + rdt) * scale) * invtess;

			if (y == 0 || y == WARP_TESS)
				dst[4] = (float) dst[1] / (float) glheight;
			else dst[4] = (t + sin (s + rdt) * scale) * invtess;

			// fuck opengl and it's fucking "bottom-left is the origin" pedantic wankery
			dst[4] = 1.0f - dst[4] * invhscale;
		}
	}

	for (y = 0, ndx = 0, i = 0; y < WARP_TESS; y++, i++)
	{
		for (x = 0; x < WARP_TESS; x++, i++, ndx += 6)
		{
			indexes[ndx + 0] = i;
			indexes[ndx + 1] = i + 1;
			indexes[ndx + 2] = i + WARP_TESS + 2;
			indexes[ndx + 3] = i;
			indexes[ndx + 4] = i + WARP_TESS + 2;
			indexes[ndx + 5] = i + WARP_TESS + 1;
		}
	}

	GL_SetIndices (0, indexes);
	GL_DrawIndexedPrimitive (GL_TRIANGLES, numindexes, (WARP_TESS + 1) * (WARP_TESS + 1));
	RTTWarp_Takedown ();

	// disable view blend because we're going to get it for free here
	//v_blend[3] = 0;
}


void GL_PolyBlend (void)
{
	float blendverts[] =
	{
		// note - 100 y is needed for high fov
		10, 100, 100,
		10, -100, 100,
		10, -100, -100,
		10, 100, -100
	};

	// whaddaya know - these run faster
	if (!gl_polyblend.value) return;
	if (!v_blend[3]) return;

	glDisable (GL_ALPHA_TEST);
	glEnable (GL_BLEND);
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_TEXTURE_2D);

	glLoadIdentity ();

	// FIXME: get rid of these
	glRotatef (-90, 1, 0, 0);	    // put Z going up
	glRotatef (90, 0, 0, 1);	    // put Z going up

	GL_SetStreamSource (0, GLSTREAM_POSITION, 3, GL_FLOAT, 0, blendverts);
	GL_SetStreamSource (0, GLSTREAM_COLOR, 0, GL_NONE, 0, NULL);
	GL_SetStreamSource (0, GLSTREAM_TEXCOORD0, 0, GL_NONE, 0, NULL);
	GL_SetStreamSource (0, GLSTREAM_TEXCOORD1, 0, GL_NONE, 0, NULL);
	GL_SetStreamSource (0, GLSTREAM_TEXCOORD2, 0, GL_NONE, 0, NULL);

	if (gl_arb_vertex_buffer_object)
		GL_SetIndices (r_quadindexbuffer, (void *) 0);
	else GL_SetIndices (0, r_quad_indexes);

	glColor4f (v_blend[0], v_blend[1], v_blend[2], v_blend[3]);

	GL_DrawIndexedPrimitive (GL_TRIANGLES, 6, 4);

	glDisable (GL_BLEND);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_ALPHA_TEST);
	glColor4f (1, 1, 1, 1);
}


void DrawAccumBlur (void)
{
	static int blurstate = 0;
	float accblur;
	static float damagetime = -1.0f;

	float r_frametime = cl.time - cl.oldtime;

	if (!r_motionblur.value) return;

	// evaluate blur conditions
	if (cl.stats[STAT_HEALTH] <= 0)
	{
		// being dead always overrides everything else
		accblur = 0.75f;
	}
	else if (cl.cshifts[CSHIFT_DAMAGE].percent)
	{
		// initial damage blur
		accblur = 0.75f;
		damagetime = 0.5f;
	}
	else if (damagetime >= 0.0f)
	{
		// persist damage blur
		damagetime -= r_frametime;
		accblur = 0.75f;
	}
	else if (cl.cshifts[CSHIFT_CONTENTS].percent)
	{
		// blur less if underwater
		accblur = 0.666f;
	}
	else accblur = -1.0f;

	if (accblur <= 0.0f)
	{
		// reinit if we're not blurring so that the contents of the
		// accumulation buffer are valid for the frame
		blurstate = 0;
		return;
	}

	if (!blurstate)
	{
		// load the scene into the accumulation buffer
		glAccum (GL_LOAD, 1.0f);
	}
	else
	{
		glAccum (GL_MULT, accblur); // scale contents of accumulation buffer
		glAccum (GL_ACCUM, 1.0f - accblur); // add screen contents
		glAccum (GL_RETURN, 1.0f); // read result back
	}

	blurstate = 1;
}


void R_EmitWirePoint (vec3_t origin)
{
	int size = 8;

	glBegin (GL_LINES);

	glVertex3f (origin[0] - size, origin[1], origin[2]);
	glVertex3f (origin[0] + size, origin[1], origin[2]);
	glVertex3f (origin[0], origin[1] - size, origin[2]);
	glVertex3f (origin[0], origin[1] + size, origin[2]);
	glVertex3f (origin[0], origin[1], origin[2] - size);
	glVertex3f (origin[0], origin[1], origin[2] + size);

	glEnd ();
}


void R_EmitWireBox (vec3_t mins, vec3_t maxs)
{
	glBegin (GL_QUAD_STRIP);

	glVertex3f (mins[0], mins[1], mins[2]);
	glVertex3f (mins[0], mins[1], maxs[2]);
	glVertex3f (maxs[0], mins[1], mins[2]);
	glVertex3f (maxs[0], mins[1], maxs[2]);
	glVertex3f (maxs[0], maxs[1], mins[2]);
	glVertex3f (maxs[0], maxs[1], maxs[2]);
	glVertex3f (mins[0], maxs[1], mins[2]);
	glVertex3f (mins[0], maxs[1], maxs[2]);
	glVertex3f (mins[0], mins[1], mins[2]);
	glVertex3f (mins[0], mins[1], maxs[2]);

	glEnd ();
}


void R_ShowBBoxes (void)
{
	int i;
	edict_t *ed;
	extern edict_t *sv_player;
	vec3_t mins, maxs;
	qboolean stateset = false;

	if (!r_showbboxes.value) return;
	if (cl.maxclients > 1) return;
	if (!sv.active) return;
	if (!r_drawentities.value) return;

	for (i = 1; i < sv.num_edicts; i++)
	{
		ed = sv.edictpointers[i];

		// don't draw player's own bbox
		if (ed == sv_player) continue;

		// don't show free edicts
		if (ed->free) continue;

		VectorAdd (ed->v.mins, ed->v.origin, mins);
		VectorAdd (ed->v.maxs, ed->v.origin, maxs);

		if (R_CullBox (mins, maxs)) continue;

		if (!stateset)
		{
			// disable vertex arrays because we're going to immediate mode
			GL_UnbindBuffers ();

			glDisable (GL_DEPTH_TEST);
			glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
			glDisable (GL_TEXTURE_2D);
			glDisable (GL_CULL_FACE);
			glColor3f (1, 1, 1);

			stateset = true;
		}

		if (ed->v.mins[0] == ed->v.maxs[0] && ed->v.mins[1] == ed->v.maxs[1] && ed->v.mins[2] == ed->v.maxs[2])
		{
			// point entity
			R_EmitWirePoint (ed->v.origin);
		}
		else
		{
			// box entity
			R_EmitWireBox (mins, maxs);
		}
	}

	if (stateset)
	{
		glColor3f (1, 1, 1);
		glEnable (GL_TEXTURE_2D);
		glEnable (GL_CULL_FACE);
		glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
		glEnable (GL_DEPTH_TEST);
	}
}


// moved - see comment in r_world.c
// void R_UpdateLightmaps (void);

void R_RenderScene (void)
{
	// mh 2010-07-16
	R_ModelSurfsBeginFrame ();
	R_AlphaListBeginFrame ();
	R_SetupScene (); // johnfitz -- this does everything that should be done once per call to RenderScene
	R_SetupGL ();
	R_SetViewport ();
	R_Clear ();
	Fog_EnableGFog ();
	R_DrawWorld ();
	R_DrawAliasModels ();
	R_DrawSpriteModels ();
	R_DrawParticles ();
	R_DrawCoronas ();
	R_DrawAlphaList ();
	R_DrawViewModel (); // johnfitz -- moved here from R_RenderView
	Fog_DisableGFog ();

	R_ShowBBoxes ();

	// moved - see comment in r_world.c
	// R_UpdateLightmaps ();

	// run motion blur
	// (i'm currently running so fast that i don;t even notice anything at all with this one)
	// DrawAccumBlur ();

	// moved polyblend back to here because it's faster
	GL_PolyBlend ();
}


/*
================
R_RenderView
================
*/
void R_RenderView (void)
{
	float	time1 = 0, time2 = 0;

	if (r_norefresh.value) return;
	if (!cl.worldmodel) Sys_Error ("R_RenderView: NULL worldmodel");
	if (r_speeds.value) time1 = Sys_FloatTime ();

	// johnfitz -- rendering statistics
	// mh - always so that we can use con_printf for debug output while developing
	rs_brushpolys = rs_aliaspolys = rs_drawelements = rs_dynamiclightmaps = 0;

	R_SetupView (); // johnfitz -- this does everything that should be done once per frame
	R_RenderScene ();

	// cleaned up code and made it much more readable here
	if (r_speeds.value)
	{
		// johnfitz -- modified r_speeds output
		// mh - moved this inside the "if"
		time2 = Sys_FloatTime ();

		Con_Printf ("%3i ms  %4i wpoly   %4i epoly   %4i draw   %3i lmap\n",
			(int) ((time2 - time1) * 1000),
			rs_brushpolys,
			rs_aliaspolys,
			rs_drawelements,
			rs_dynamiclightmaps);
	}
	// johnfitz
}


