/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//gl_texmgr.c -- fitzquake's texture manager. manages opengl texture images

// this whole file is REALLY STARTING TO GET ON MY NERVES
// it doesn't cache textures between maps
// it only works with GL_TEXTURE_2D
// it's retained a hard-limit of 2048 textures
// the naming convention and coding style makes it difficult for me to read
// it still assumes that 3-component textures actually exist in hardware

// mh 2010-07-16 rmq version
#include "quakedef.h"

extern qboolean	gl_ext_texture_filter_anisotropic;

cvar_t		gl_texture_anisotropy = {"gl_texture_anisotropy", "1", true};
cvar_t		gl_max_size = {"gl_max_size", "0"};
cvar_t		gl_picmip = {"gl_picmip", "0"};
GLint			gl_hardware_maxsize;

gltexture_t	*active_gltextures, *free_gltextures;
gltexture_t	*notexture, *nulltexture;
int numgltextures;

unsigned int d_8to24table[256];
unsigned int d_8to24table_rgba[256];
unsigned int d_8to24table_fbright[256];
unsigned int d_8to24table_nobright[256];
unsigned int d_8to24table_conchars[256];
unsigned int d_8to24table_shirt[256];
unsigned int d_8to24table_pants[256];

/*
================================================================================

	COMMANDS

================================================================================
*/

typedef struct
{
	int	magfilter;
	int minfilter;
	char *name;
} glmode_t;
glmode_t modes[] =
{
	{GL_NEAREST, GL_NEAREST,				"GL_NEAREST"},
	{GL_NEAREST, GL_NEAREST_MIPMAP_NEAREST,	"GL_NEAREST_MIPMAP_NEAREST"},
	{GL_NEAREST, GL_NEAREST_MIPMAP_LINEAR,	"GL_NEAREST_MIPMAP_LINEAR"},
	{GL_LINEAR,  GL_LINEAR,					"GL_LINEAR"},
	{GL_LINEAR,  GL_LINEAR_MIPMAP_NEAREST,	"GL_LINEAR_MIPMAP_NEAREST"},
	{GL_LINEAR,  GL_LINEAR_MIPMAP_LINEAR,	"GL_LINEAR_MIPMAP_LINEAR"},
};
#define NUM_GLMODES 6
int gl_texturemode = 5; // bilinear

void GL_PixelStore (int rowalign, int rowlen)
{
	static int oldrowalign = -1;
	static int oldrowlen = -1;

	if (oldrowalign != rowalign)
	{
		glPixelStorei (GL_UNPACK_ALIGNMENT, rowalign);
		oldrowalign = rowalign;
	}

	if (oldrowlen != rowlen)
	{
		glPixelStorei (GL_UNPACK_ROW_LENGTH, rowlen);
		oldrowlen = rowlen;
	}
}


/*
===============
TM_DescribeTextureModes_f -- report available texturemodes
===============
*/
void TM_DescribeTextureModes_f (void)
{
	int i;

	for (i = 0; i < NUM_GLMODES; i++)
		Con_SafePrintf ("   %2i: %s\n", i + 1, modes[i].name);

	Con_Printf ("%i modes\n", i);
}

/*
===============
TM_SetFilterModes
===============
*/
void TM_SetFilterModes (gltexture_t *glt)
{
	GL_BindTexture (GL_TEXTURE0, glt);

	if (glt->flags & TEXPREF_NEAREST)
	{
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	}
	else if (glt->flags & TEXPREF_LINEAR)
	{
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	}
	else if (glt->flags & TEXPREF_MIPMAP)
	{
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, modes[gl_texturemode].magfilter);
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, modes[gl_texturemode].minfilter);
	}
	else
	{
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, modes[gl_texturemode].magfilter);
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, modes[gl_texturemode].magfilter);
	}

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, gl_texture_anisotropy.value);
}

/*
===============
TM_TextureMode_f
===============
*/
void TM_TextureMode_f (void)
{
	gltexture_t	*glt;
	char *arg;
	int i;

	switch (Cmd_Argc())
	{
	case 1:
		Con_Printf ("\"gl_texturemode\" is \"%s\"\n", modes[gl_texturemode].name);
		break;
	case 2:
		arg = Cmd_Argv (1);

		if (arg[0] == 'G' || arg[0] == 'g')
		{
			for (i = 0; i < NUM_GLMODES; i++)
			{
				if (!Q_strcasecmp (modes[i].name, arg))
				{
					gl_texturemode = i;
					goto stuff;
				}
			}

			Con_Printf ("\"%s\" is not a valid texturemode\n", arg);
			return;
		}
		else if (arg[0] >= '0' && arg[0] <= '9')
		{
			i = atoi (arg);

			if (i > NUM_GLMODES || i < 1)
			{
				Con_Printf ("\"%s\" is not a valid texturemode\n", arg);
				return;
			}

			gl_texturemode = i - 1;
		}
		else
			Con_Printf ("\"%s\" is not a valid texturemode\n", arg);

stuff:

		for (glt = active_gltextures; glt; glt = glt->next)
			TM_SetFilterModes (glt);

		break;
	default:
		Con_SafePrintf ("usage: gl_texturemode <mode>\n");
		break;
	}
}

/*
===============
TM_Anisotropy_f -- called when gl_texture_anisotropy changes

FIXME: this is getting called twice (becuase of the recursive Cvar_SetValue call)
===============
*/
void TM_Anisotropy_f (void)
{
	extern float gl_max_anisotropy;
	gltexture_t	*glt;

	Cvar_SetValue ("gl_texture_anisotropy", CLAMP (1.0f, gl_texture_anisotropy.value, gl_max_anisotropy));

	for (glt = active_gltextures; glt; glt = glt->next)
		TM_SetFilterModes (glt);
}

/*
===============
TM_Imagelist_f -- report loaded textures
===============
*/
void TM_Imagelist_f (void)
{
	float mb;
	float texels = 0;
	gltexture_t	*glt;
	extern cvar_t vid_bpp;

	for (glt = active_gltextures; glt; glt = glt->next)
	{
		Con_SafePrintf ("   %4i x%4i %s\n", glt->width, glt->height, glt->name);
		texels += (glt->width * glt->height);
	}

	mb = texels * (vid_bpp.value / 8.0f) / 0x100000;
	Con_Printf ("%i textures %i pixels %1.1f megabytes\n", numgltextures, (int) texels, mb);
}

/*
===============
TM_Imagedump_f -- dump all current textures to TGA files
===============
*/
void TM_Imagedump_f (void)
{
	char tganame[MAX_OSPATH], tempname[MAX_OSPATH], dirname[MAX_OSPATH];
	gltexture_t	*glt;
	byte *buffer;
	char *c;

	//create directory
	sprintf (dirname, "%s/imagedump", com_gamedir);
	Sys_mkdir (dirname);

	//loop through textures
	for (glt = active_gltextures; glt; glt = glt->next)
	{
		Q_strcpy (tempname, glt->name);

		while ((c = strchr (tempname, ':'))) * c = '_';
		while ((c = strchr (tempname, '/'))) * c = '_';
		while ((c = strchr (tempname, '*'))) * c = '_';

		sprintf (tganame, "imagedump/%s.tga", tempname);

		GL_BindTexture (GL_TEXTURE0, glt);

		buffer = (byte *) malloc (glt->width * glt->height * 4);
		glGetTexImage (GL_TEXTURE_2D, 0, GL_BGRA, GL_UNSIGNED_BYTE, buffer);
		Image_WriteTGA (tganame, buffer, glt->width, glt->height, 32, true);

		free (buffer);
	}

	Con_Printf ("dumped %i textures to %s\n", numgltextures, dirname);
}

/*
===============
TM_FrameUsage -- report texture memory usage for this frame
===============
*/
float TM_FrameUsage (void)
{
	float mb;
	float texels = 0;
	gltexture_t	*glt;
	extern cvar_t vid_bpp;

	for (glt = active_gltextures; glt; glt = glt->next)
	{
		if (glt->visframe == r_framecount)
		{
			texels += (glt->width * glt->height);
		}
	}

	mb = texels * (vid_bpp.value / 8.0f) / 0x100000;
	return mb;
}

/*
================================================================================

	TEXTURE MANAGER

================================================================================
*/

/*
================
TM_FindTexture
================
*/
gltexture_t *TM_FindTexture (model_t *owner, char *name)
{
	gltexture_t	*glt;

	if (name)
	{
		for (glt = active_gltextures; glt; glt = glt->next)
			if (glt->owner == owner && !strcmp (glt->name, name))
				return glt;
	}

	return NULL;
}

/*
================
TM_NewTexture
================
*/
gltexture_t *TM_NewTexture (void)
{
	gltexture_t *glt;

	if (!free_gltextures)
	{
		free_gltextures = (gltexture_t *) malloc (sizeof (gltexture_t));
		memset (free_gltextures, 0, sizeof (gltexture_t));
		free_gltextures->next = NULL;
	}

	glt = free_gltextures;
	free_gltextures = glt->next;
	glt->next = active_gltextures;
	active_gltextures = glt;

	glGenTextures (1, (GLuint *) &glt->texnum);
	numgltextures++;
	return glt;
}


/*
================
TM_FreeTexture
================
*/
void TM_FreeTexture (gltexture_t *kill)
{
	gltexture_t *glt;

	if (kill == NULL)
	{
		Con_Printf ("TM_FreeTexture: NULL texture\n");
		return;
	}

	if (active_gltextures == kill)
	{
		active_gltextures = kill->next;
		kill->next = free_gltextures;
		free_gltextures = kill;

		glDeleteTextures (1, (const GLuint *) &kill->texnum);
		numgltextures--;
		return;
	}

	for (glt = active_gltextures; glt; glt = glt->next)
	{
		if (glt->next == kill)
		{
			glt->next = kill->next;
			kill->next = free_gltextures;
			free_gltextures = kill;

			glDeleteTextures (1, (const GLuint *) &kill->texnum);
			numgltextures--;
			return;
		}
	}

	Con_DPrintf ("TM_FreeTexture: not found\n");
}

/*
================
TM_FreeTextures

compares each bit in "flags" to the one in glt->flags only if that bit is active in "mask"
================
*/
void TM_FreeTextures (int flags, int mask)
{
	gltexture_t *glt, *next;

	for (glt = active_gltextures; glt; glt = next)
	{
		next = glt->next;

		if ((glt->flags & mask) == (flags & mask))
			TM_FreeTexture (glt);
	}
}


void TM_DeleteTextureObjects (void)
{
	gltexture_t *glt;

	for (glt = active_gltextures; glt; glt = glt->next)
	{
		glDeleteTextures (1, &glt->texnum);
		glt->texnum = 0;
	}
}


/*
================
TM_FreeTexturesForOwner
================
*/
void TM_FreeTexturesForOwner (model_t *owner)
{
	gltexture_t *glt, *next;

	if (!owner) goto free_unowned;

	switch (owner->type)
	{
	case mod_alias:
	case mod_brush:
	case mod_sprite:
		// Con_DPrintf ("freeing textures for %s\n", owner->name);
free_unowned:;
		for (glt = active_gltextures; glt; glt = next)
		{
			next = glt->next;

			if (glt && glt->owner == owner)
				TM_FreeTexture (glt);
		}

		return;

	default:
		break;
	}
}

/*
================================================================================

	INIT

================================================================================
*/

/*
=================
TM_LoadPalette -- johnfitz -- was VID_SetPalette, moved here, renamed, rewritten
=================
*/
void TM_LoadPalette (void)
{
	byte *pal, *src, *dst;
	int i, mark;
	FILE *f;

	COM_FOpenFile ("gfx/palette.lmp", &f);

	if (!f)
		Sys_Error ("Couldn't load gfx/palette.lmp");

	mark = Hunk_LowMark ();
	pal = (byte *) Hunk_Alloc (768);
	i = fread (pal, 1, 768, f);
	fclose (f);

	// standard palette, 255 is transparent
	dst = (byte *) d_8to24table;
	src = pal;

	for (i = 0; i < 256; i++)
	{
		dst[2] = *src++;
		dst[1] = *src++;
		dst[0] = *src++;
		dst[3] = 255;
		dst += 4;
	}

	// prevent pink fringes by setting alpha to black
	d_8to24table[255] = 0;

	// rgba table for particles and anything else that needs to index the palette explicitly
	dst = (byte *) d_8to24table_rgba;
	src = pal;

	for (i = 0; i < 256; i++)
	{
		dst[0] = *src++;
		dst[1] = *src++;
		dst[2] = *src++;
		dst[3] = 255;
		dst += 4;
	}

	// prevent pink fringes by setting alpha to black
	d_8to24table[255] = 0;

	// fullbright palette, 0-223 are black (for additive blending)
	src = pal + 224 * 3;
	dst = (byte *) (d_8to24table_fbright) + 224 * 4;

	for (i = 224; i < 256; i++)
	{
		dst[2] = *src++;
		dst[1] = *src++;
		dst[0] = *src++;
		dst[3] = 255;
		dst += 4;
	}

	for (i = 0; i < 224; i++)
	{
		// comply with strict aliasing
		byte *rgba = (byte *) &d_8to24table_fbright[i];

		rgba[0] = 0;
		rgba[1] = 0;
		rgba[2] = 0;
		rgba[3] = 255;
	}

	//nobright palette, 224-255 are black (for additive blending)
	dst = (byte *) d_8to24table_nobright;
	src = pal;

	for (i = 0; i < 256; i++)
	{
		dst[2] = *src++;
		dst[1] = *src++;
		dst[0] = *src++;
		dst[3] = 255;
		dst += 4;
	}

	for (i = 224; i < 256; i++)
	{
		// comply with strict aliasing
		byte *rgba = (byte *) &d_8to24table_nobright[i];

		rgba[0] = 0;
		rgba[1] = 0;
		rgba[2] = 0;
		rgba[3] = 255;
	}

	// correct alpha colour
	d_8to24table_nobright[255] = 0;
	d_8to24table_fbright[255] = 0;

	//conchars palette, 0 and 255 are transparent
	memcpy (d_8to24table_conchars, d_8to24table, 256 * 4);
	d_8to24table_conchars[0] = 0;

	Hunk_FreeToLowMark (mark);
}

/*
================
TM_NewGame
================
*/
void TM_NewGame (void)
{
	TM_FreeTextures (0, TEXPREF_PERSIST); //deletes all textures where TEXPREF_PERSIST is unset
	TM_LoadPalette ();
}


/*
================
TM_Init

must be called before any texture loading
================
*/
void TM_Init (void)
{
	static byte notexture_data[16] = {159, 91, 83, 255, 0, 0, 0, 255, 0, 0, 0, 255, 159, 91, 83, 255}; //black and pink checker
	static byte nulltexture_data[16] = {127, 191, 255, 255, 0, 0, 0, 255, 0, 0, 0, 255, 127, 191, 255, 255}; //black and blue checker
	extern texture_t *r_notexture_mip, *r_notexture_mip2;

	// init texture list
	free_gltextures = NULL;
	active_gltextures = NULL;
	numgltextures = 0;

	// palette
	TM_LoadPalette ();

	Cvar_RegisterVariable (&gl_max_size, NULL);
	Cvar_RegisterVariable (&gl_picmip, NULL);
	Cvar_RegisterVariable (&gl_texture_anisotropy, &TM_Anisotropy_f);
	Cmd_AddCommand ("gl_texturemode", &TM_TextureMode_f);
	Cmd_AddCommand ("gl_describetexturemodes", &TM_DescribeTextureModes_f);
	Cmd_AddCommand ("imagelist", &TM_Imagelist_f);
	Cmd_AddCommand ("imagedump", &TM_Imagedump_f);

	// poll max size from hardware
	glGetIntegerv (GL_MAX_TEXTURE_SIZE, &gl_hardware_maxsize);

	// load notexture images
	notexture = TM_LoadImage (NULL, "notexture", 2, 2, SRC_RGBA, notexture_data, "", (src_offset_t) notexture_data, TEXPREF_NEAREST | TEXPREF_PERSIST | TEXPREF_NOPICMIP);
	nulltexture = TM_LoadImage (NULL, "nulltexture", 2, 2, SRC_RGBA, nulltexture_data, "", (src_offset_t) nulltexture_data, TEXPREF_NEAREST | TEXPREF_PERSIST | TEXPREF_NOPICMIP);

	//have to assign these here becuase Mod_Init is called before TM_Init
	r_notexture_mip->gltexture = r_notexture_mip2->gltexture = notexture;
}


/*
================================================================================

	IMAGE LOADING

================================================================================
*/

/*
================
TM_Pad -- return smallest power of two greater than or equal to s
================
*/
int TM_Pad (int s)
{
	int i;

	for (i = 1; i < s; i <<= 1);

	return i;
}

/*
===============
TM_SafeTextureSize -- return a size with hardware and user prefs in mind
===============
*/
int TM_SafeTextureSize (int s)
{
	s = TM_Pad (s);

	if ((int) gl_max_size.value > 0)
		s = min (TM_Pad ((int) gl_max_size.value), s);

	s = min (gl_hardware_maxsize, s);
	return s;
}

/*
================
TM_PadConditional -- only pad if a texture of that size would be padded. (used for tex coords)
================
*/
int TM_PadConditional (int s)
{
	if (s < TM_SafeTextureSize (s))
		return TM_Pad (s);
	else
		return s;
}

/*
================
TM_MipMapW
================
*/
unsigned *TM_MipMapW (unsigned *data, int width, int height)
{
	int		i, size;
	byte	*out, *in;

	out = in = (byte *) data;
	size = (width * height) >> 1;

	for (i = 0; i < size; i++, out += 4, in += 8)
	{
		out[0] = (in[0] + in[4]) >> 1;
		out[1] = (in[1] + in[5]) >> 1;
		out[2] = (in[2] + in[6]) >> 1;
		out[3] = (in[3] + in[7]) >> 1;
	}

	return data;
}

/*
================
TM_MipMapH
================
*/
unsigned *TM_MipMapH (unsigned *data, int width, int height)
{
	int		i, j;
	byte	*out, *in;

	out = in = (byte *) data;
	height >>= 1;
	width <<= 2;

	for (i = 0; i < height; i++, in += width)
	{
		for (j = 0; j < width; j += 4, out += 4, in += 4)
		{
			out[0] = (in[0] + in[width+0]) >> 1;
			out[1] = (in[1] + in[width+1]) >> 1;
			out[2] = (in[2] + in[width+2]) >> 1;
			out[3] = (in[3] + in[width+3]) >> 1;
		}
	}

	return data;
}

/*
================
TM_ResampleTexture -- bilinear resample
================
*/
unsigned *TM_ResampleTexture (unsigned *in, int inwidth, int inheight, qboolean alpha)
{
	byte *nwpx, *nepx, *swpx, *sepx, *dest;
	unsigned xfrac, yfrac, x, y, modx, mody, imodx, imody, injump, outjump;
	unsigned *out;
	int i, j, outwidth, outheight;

	if (inwidth == TM_Pad (inwidth) && inheight == TM_Pad (inheight))
		return in;

	outwidth = TM_Pad (inwidth);
	outheight = TM_Pad (inheight);
	out = (unsigned *) Hunk_Alloc (outwidth * outheight * 4);

	xfrac = ((inwidth - 1) << 16) / (outwidth - 1);
	yfrac = ((inheight - 1) << 16) / (outheight - 1);
	y = outjump = 0;

	for (i = 0; i < outheight; i++)
	{
		mody = (y >> 8) & 0xFF;
		imody = 256 - mody;
		injump = (y >> 16) * inwidth;
		x = 0;

		for (j = 0; j < outwidth; j++)
		{
			modx = (x >> 8) & 0xFF;
			imodx = 256 - modx;

			nwpx = (byte *) (in + (x >> 16) + injump);
			nepx = nwpx + 4;
			swpx = nwpx + inwidth * 4;
			sepx = swpx + 4;

			dest = (byte *) (out + outjump + j);

			dest[0] = (nwpx[0] * imodx * imody + nepx[0] * modx * imody + swpx[0] * imodx * mody + sepx[0] * modx * mody) >> 16;
			dest[1] = (nwpx[1] * imodx * imody + nepx[1] * modx * imody + swpx[1] * imodx * mody + sepx[1] * modx * mody) >> 16;
			dest[2] = (nwpx[2] * imodx * imody + nepx[2] * modx * imody + swpx[2] * imodx * mody + sepx[2] * modx * mody) >> 16;

			if (alpha)
				dest[3] = (nwpx[3] * imodx * imody + nepx[3] * modx * imody + swpx[3] * imodx * mody + sepx[3] * modx * mody) >> 16;
			else
				dest[3] = 255;

			x += xfrac;
		}

		outjump += outwidth;
		y += yfrac;
	}

	return out;
}

/*
===============
TM_AlphaEdgeFix

eliminate pink edges on sprites, etc.
operates in place on 32bit data
===============
*/
void TM_AlphaEdgeFix (byte *data, int width, int height)
{
	int i, j, n = 0, b, c[3] = {0, 0, 0}, lastrow, thisrow, nextrow, lastpix, thispix, nextpix;
	byte *dest = data;

	for (i = 0; i < height; i++)
	{
		lastrow = width * 4 * ((i == 0) ? height - 1 : i - 1);
		thisrow = width * 4 * i;
		nextrow = width * 4 * ((i == height - 1) ? 0 : i + 1);

		for (j = 0; j < width; j++, dest += 4)
		{
			if (dest[3]) //not transparent
				continue;

			lastpix = 4 * ((j == 0) ? width - 1 : j - 1);
			thispix = 4 * j;
			nextpix = 4 * ((j == width - 1) ? 0 : j + 1);

			b = lastrow + lastpix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}
			b = thisrow + lastpix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}
			b = nextrow + lastpix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}
			b = lastrow + thispix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}
			b = nextrow + thispix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}
			b = lastrow + nextpix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}
			b = thisrow + nextpix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}
			b = nextrow + nextpix; if (data[b+3]) {c[0] += data[b]; c[1] += data[b+1]; c[2] += data[b+2]; n++;}

			//average all non-transparent neighbors
			if (n)
			{
				dest[0] = (byte) (c[0] / n);
				dest[1] = (byte) (c[1] / n);
				dest[2] = (byte) (c[2] / n);

				n = c[0] = c[1] = c[2] = 0;
			}
		}
	}
}

/*
===============
TM_PadEdgeFixW -- special case of AlphaEdgeFix for textures that only need it because they were padded

operates in place on 32bit data, and expects unpadded height and width values
===============
*/
void TM_PadEdgeFixW (byte *data, int width, int height)
{
	byte *src, *dst;
	int i, padw, padh;

	padw = TM_PadConditional (width);
	padh = TM_PadConditional (height);

	//copy last full column to first empty column, leaving alpha byte at zero
	src = data + (width - 1) * 4;

	for (i = 0; i < padh; i++)
	{
		src[4] = src[0];
		src[5] = src[1];
		src[6] = src[2];
		src += padw * 4;
	}

	//copy first full column to last empty column, leaving alpha byte at zero
	src = data;
	dst = data + (padw - 1) * 4;

	for (i = 0; i < padh; i++)
	{
		dst[0] = src[0];
		dst[1] = src[1];
		dst[2] = src[2];
		src += padw * 4;
		dst += padw * 4;
	}
}

/*
===============
TM_PadEdgeFixH -- special case of AlphaEdgeFix for textures that only need it because they were padded

operates in place on 32bit data, and expects unpadded height and width values
===============
*/
void TM_PadEdgeFixH (byte *data, int width, int height)
{
	byte *src, *dst;
	int i, padw, padh;

	padw = TM_PadConditional (width);
	padh = TM_PadConditional (height);

	//copy last full row to first empty row, leaving alpha byte at zero
	dst = data + height * padw * 4;
	src = dst - padw * 4;

	for (i = 0; i < padw; i++)
	{
		dst[0] = src[0];
		dst[1] = src[1];
		dst[2] = src[2];
		src += 4;
		dst += 4;
	}

	//copy first full row to last empty row, leaving alpha byte at zero
	dst = data + (padh - 1) * padw * 4;
	src = data;

	for (i = 0; i < padw; i++)
	{
		dst[0] = src[0];
		dst[1] = src[1];
		dst[2] = src[2];
		src += 4;
		dst += 4;
	}
}

/*
================
TM_8to32
================
*/
unsigned *TM_8to32 (byte *in, int pixels, unsigned int *usepal)
{
	int i;
	unsigned *out, *data;

	out = data = (unsigned *) Hunk_Alloc (pixels * 4);

	for (i = 0; i < pixels; i++)
		*out++ = usepal[*in++];

	return data;
}

/*
================
TM_PadImageW -- return image with width padded up to power-of-two dimentions
================
*/
byte *TM_PadImageW (byte *in, int width, int height, byte padbyte)
{
	int i, j, outwidth;
	byte *out, *data;

	if (width == TM_Pad (width))
		return in;

	outwidth = TM_Pad (width);

	out = data = (byte *) Hunk_Alloc (outwidth * height);

	for (i = 0; i < height; i++)
	{
		for (j = 0; j < width; j++)
			*out++ = *in++;

		for (; j < outwidth; j++)
			*out++ = padbyte;
	}

	return data;
}

/*
================
TM_PadImageH -- return image with height padded up to power-of-two dimentions
================
*/
byte *TM_PadImageH (byte *in, int width, int height, byte padbyte)
{
	int i, srcpix, dstpix;
	byte *data, *out;

	if (height == TM_Pad (height))
		return in;

	srcpix = width * height;
	dstpix = width * TM_Pad (height);

	out = data = (byte *) Hunk_Alloc (dstpix);

	for (i = 0; i < srcpix; i++)
		*out++ = *in++;

	for (; i < dstpix; i++)
		*out++ = padbyte;

	return data;
}

void TM_Resample8BitTexture (byte *in, int inwidth, int inheight, unsigned *out, int outwidth, int outheight, unsigned *table)
{
	int		i, j;
	byte	*inrow, *inrow2;
	unsigned	frac, fracstep;
	unsigned	*p1, *p2;
	byte		*pix1, *pix2, *pix3, *pix4;

	p1 = (unsigned *) Hunk_Alloc (outwidth * 4);
	p2 = (unsigned *) Hunk_Alloc (outwidth * 4);

	fracstep = inwidth * 0x10000 / outwidth;
	frac = fracstep >> 2;

	for (i = 0; i < outwidth; i++)
	{
		p1[i] = (frac >> 16);
		frac += fracstep;
	}

	frac = 3 * (fracstep >> 2);

	for (i = 0; i < outwidth; i++)
	{
		p2[i] = (frac >> 16);
		frac += fracstep;
	}

	for (i = 0; i < outheight; i++, out += outwidth)
	{
		inrow = in + inwidth * (int) ((i + 0.25) * inheight / outheight);
		inrow2 = in + inwidth * (int) ((i + 0.75) * inheight / outheight);
		frac = fracstep >> 1;

		for (j = 0; j < outwidth; j++)
		{
			pix1 = (byte *) &table[inrow[p1[j]]];
			pix2 = (byte *) &table[inrow[p2[j]]];
			pix3 = (byte *) &table[inrow2[p1[j]]];
			pix4 = (byte *) &table[inrow2[p2[j]]];

			((byte *) (out + j)) [0] = (pix1[0] + pix2[0] + pix3[0] + pix4[0]) >> 2;
			((byte *) (out + j)) [1] = (pix1[1] + pix2[1] + pix3[1] + pix4[1]) >> 2;
			((byte *) (out + j)) [2] = (pix1[2] + pix2[2] + pix3[2] + pix4[2]) >> 2;
			((byte *) (out + j)) [3] = (pix1[3] + pix2[3] + pix3[3] + pix4[3]) >> 2;
		}
	}
}


/*
================
TM_LoadImage32 -- handles 32bit source data
================
*/
void TM_LoadImage32 (gltexture_t *glt, unsigned *data)
{
	int	miplevel, mipwidth, mipheight, picmip;

	// resample up
	data = TM_ResampleTexture (data, glt->width, glt->height, glt->flags & TEXPREF_ALPHA);
	glt->width = TM_Pad (glt->width);
	glt->height = TM_Pad (glt->height);

	// mipmap down
	picmip = (glt->flags & TEXPREF_NOPICMIP) ? 0 : max ((int) gl_picmip.value, 0);
	mipwidth = TM_SafeTextureSize (glt->width >> picmip);
	mipheight = TM_SafeTextureSize (glt->height >> picmip);

	while (glt->width > mipwidth)
	{
		TM_MipMapW (data, glt->width, glt->height);
		glt->width >>= 1;

		if (glt->flags & TEXPREF_ALPHA)
			TM_AlphaEdgeFix ((byte *) data, glt->width, glt->height);
	}

	while (glt->height > mipheight)
	{
		TM_MipMapH (data, glt->width, glt->height);
		glt->height >>= 1;

		if (glt->flags & TEXPREF_ALPHA)
			TM_AlphaEdgeFix ((byte *) data, glt->width, glt->height);
	}

	// upload
	GL_BindTexture (GL_TEXTURE0, glt);
	glEnable (GL_TEXTURE_2D);	// some ATI drivers might need this... http://www.opengl.org/wiki/Common_Mistakes#Automatic_mipmap_generation
	GL_PixelStore (4, 0);

	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, glt->width, glt->height, 0, GL_BGRA, GL_UNSIGNED_BYTE, data);

	// upload mipmaps
	if (glt->flags & TEXPREF_MIPMAP)
	{
		mipwidth = glt->width;
		mipheight = glt->height;

		for (miplevel = 1; mipwidth > 1 || mipheight > 1; miplevel++)
		{
			if (mipwidth > 1)
			{
				TM_MipMapW (data, mipwidth, mipheight);
				mipwidth >>= 1;
			}

			if (mipheight > 1)
			{
				TM_MipMapH (data, mipwidth, mipheight);
				mipheight >>= 1;
			}

			glTexImage2D (GL_TEXTURE_2D, miplevel, GL_RGBA, mipwidth, mipheight, 0, GL_BGRA, GL_UNSIGNED_BYTE, data);
		}
	}

	// set filter modes
	TM_SetFilterModes (glt);
}

/*
================
TM_LoadImage8 -- handles 8bit source data, then passes it to LoadImage32
================
*/
void TM_LoadImage8 (gltexture_t *glt, byte *data)
{
	extern cvar_t gl_fullbrights;
	qboolean padw = false, padh = false;
	byte padbyte;
	unsigned int *usepal;
	int i;
	int scaled_width, scaled_height;

	// HACK HACK HACK -- taken from tomazquake
	if (Q_strstr (glt->name, "shot1sid") && glt->width == 32 && glt->height == 32 && CRC_Block (data, 1024) == 65393)
	{
		// This texture in b_shell1.bsp has some of the first 32 pixels painted white.
		// They are invisible in software, but look really ugly in GL. So we just copy
		// 32 pixels from the bottom to make it look nice.
		memcpy (data, data + 32 * 31, 32);
	}

	// detect false alpha cases
	// note that this might be bollocks because OpenGL is going to load it as 4 component anyway
	if (glt->flags & TEXPREF_ALPHA && !(glt->flags & TEXPREF_CONCHARS))
	{
		for (i = 0; i < glt->width * glt->height; i++)
			if (data[i] == 255) //transparent index
				break;

		if (i == glt->width * glt->height)
			glt->flags -= TEXPREF_ALPHA;
	}

	// choose palette and padbyte
	if (glt->flags & TEXPREF_FULLBRIGHT)
	{
		usepal = d_8to24table_fbright;
		padbyte = 0;
	}
	else if (glt->flags & TEXPREF_NOBRIGHT && gl_fullbrights.value)
	{
		usepal = d_8to24table_nobright;
		padbyte = 0;
	}
	else if (glt->flags & TEXPREF_CONCHARS)
	{
		usepal = d_8to24table_conchars;
		padbyte = 0;
	}
	else
	{
		usepal = d_8to24table;
		padbyte = 255;
	}

	// pad each dimention, but only if it's not going to be downsampled later
	if (glt->flags & TEXPREF_PAD)
	{
		if (glt->width < TM_SafeTextureSize (glt->width))
		{
			data = TM_PadImageW (data, glt->width, glt->height, padbyte);
			glt->width = TM_Pad (glt->width);
			padw = true;
		}

		if (glt->height < TM_SafeTextureSize (glt->height))
		{
			data = TM_PadImageH (data, glt->width, glt->height, padbyte);
			glt->height = TM_Pad (glt->height);
			padh = true;
		}
	}

	// check for resampling and resample/upsample together here
	scaled_width = TM_Pad (glt->width);
	scaled_height = TM_Pad (glt->height);

	if (glt->width == scaled_width && glt->height == scaled_height)
	{
		// convert to 32bit
		data = (byte *) TM_8to32 (data, glt->width * glt->height, usepal);
	}
	else
	{
		// resample and upsample in-place
		// this resolves a LOT of the slowdowns when loading maps here
		// and also relieves memory pressure significantly
		unsigned *scaled = (unsigned *) Hunk_Alloc (scaled_width * scaled_height * 4);

		TM_Resample8BitTexture (data, glt->width, glt->height, scaled, scaled_width, scaled_height, usepal);

		glt->width = scaled_width;
		glt->height = scaled_height;
		data = (byte *) scaled;
	}

	// fix edges
	if (glt->flags & TEXPREF_ALPHA)
		TM_AlphaEdgeFix (data, glt->width, glt->height);
	else
	{
		if (padw) TM_PadEdgeFixW (data, glt->source_width, glt->source_height);
		if (padh) TM_PadEdgeFixH (data, glt->source_width, glt->source_height);
	}

	// upload it
	TM_LoadImage32 (glt, (unsigned *) data);
}

/*
================
TM_LoadLightmap -- handles lightmap data
================
*/
void TM_LoadLightmap (gltexture_t *glt, byte *data)
{
	extern GLenum gl_Lightmap_Type;
	extern GLenum gl_Lightmap_Format;

	// upload it
	GL_BindTexture (GL_TEXTURE0, glt);

	// internal format can't be bgra but format can be
	GL_PixelStore (4, 0);

	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, glt->width, glt->height, 0, gl_Lightmap_Format, gl_Lightmap_Type, data);

	// set filter modes
	TM_SetFilterModes (glt);
}

/*
================
TM_LoadImage -- the one entry point for loading all textures
================
*/
gltexture_t *TM_LoadImage (model_t *owner, char *name, int width, int height, enum srcformat format,
							   byte *data, char *source_file, src_offset_t source_offset, unsigned flags)
{
	unsigned short crc = 0;
	gltexture_t *glt = NULL;
	int mark = 0;

	if (isDedicated)
		return NULL;

	// cache check
	switch (format)
	{
	case SRC_INDEXED:
		crc = CRC_Block (data, width * height);
		break;
	case SRC_LIGHTMAP:
		crc = CRC_Block (data, width * height * 4);
		break;
	case SRC_RGBA:
		crc = CRC_Block (data, width * height * 4);
		break;
	}

	if ((flags & TEXPREF_OVERWRITE) && (glt = TM_FindTexture (owner, name)))
	{
		if (glt->source_crc == crc)
		{
			// Con_Printf ("reused texture %s\n", glt->name);
			return glt;
		}
	}
	else glt = TM_NewTexture ();

	// copy data
	glt->owner = owner;
	strncpy (glt->name, name, sizeof (glt->name));
	glt->width = width;
	glt->height = height;
	glt->flags = flags;
	glt->shirt = -1;
	glt->pants = -1;
	strncpy (glt->source_file, source_file, sizeof (glt->source_file));
	glt->source_offset = source_offset;
	glt->source_format = format;
	glt->source_width = width;
	glt->source_height = height;
	glt->source_crc = crc;

	//upload it
	mark = Hunk_LowMark();

	switch (glt->source_format)
	{
	case SRC_INDEXED:
		TM_LoadImage8 (glt, data);
		break;
	case SRC_LIGHTMAP:
		TM_LoadLightmap (glt, data);
		break;
	case SRC_RGBA:
		TM_LoadImage32 (glt, (unsigned *) data);
		break;
	}

	Hunk_FreeToLowMark (mark);

	return glt;
}

/*
================================================================================

	COLORMAPPING AND TEXTURE RELOADING

================================================================================
*/

/*
================
TM_ReloadImage -- reloads a texture, and colormaps it if needed
================
*/
void Mod_FloodFillSkin (byte *skin, int skinwidth, int skinheight);

void TM_ReloadImage (gltexture_t *glt, int shirt, int pants)
{
	byte	translation[256];
	byte	*src, *dst, *data = NULL, *translated;
	int		mark, size, i;
	FILE	*f = NULL;

	// get source data
	mark = Hunk_LowMark ();

	if (glt->source_file[0] && glt->source_offset)
	{
		// lump inside file
#if 0
		// this is HORRIBLE because it needs to reload THE ENTIRE FUCKING FILE into memory
		// it can also crash if a really large source file is loaded and the hunk overflows
		// plus it's about 20 times slower than it need be
		data = COM_LoadHunkFile (glt->source_file);

		if (!data)
			goto invalid;

		data += glt->source_offset;
#else
		// aaaaah, much better
		COM_FOpenFile (glt->source_file, &f);

		if (!f) goto invalid;

		data = Hunk_Alloc (glt->source_width * glt->source_height * (glt->source_format == SRC_INDEXED ? 1 : 4));

		// need SEEK_CUR for PAKs and WADs
		fseek (f, glt->source_offset, SEEK_CUR);
		fread (data, glt->source_width * glt->source_height * (glt->source_format == SRC_INDEXED ? 1 : 4), 1, f);

		// we must null this so that it doesn't try to fclose twice
		fclose (f);
		f = NULL;
#endif
	}
	else if (glt->source_file[0] && !glt->source_offset)
		data = Image_LoadImage (glt->source_file, (int *) &glt->source_width, (int *) &glt->source_height); //simple file
	else if (!glt->source_file[0] && glt->source_offset)
		data = (byte *) glt->source_offset; //image in memory

	if (!data)
	{
invalid:;
		Con_Printf ("TM_ReloadImage: invalid source for %s\n", glt->name);
		Hunk_FreeToLowMark (mark);
		if (f) fclose (f);
		return;
	}

	glt->width = glt->source_width;
	glt->height = glt->source_height;

	// apply shirt and pants colors
	// if shirt and pants are -1,-1, use existing shirt and pants colors
	// if existing shirt and pants colors are -1,-1, don't bother colormapping
	if (shirt > -1 && pants > -1)
	{
		if (glt->source_format == SRC_INDEXED)
		{
			glt->shirt = shirt;
			glt->pants = pants;
		}
		else
			Con_Printf ("TM_ReloadImage: can't colormap a non SRC_INDEXED texture: %s\n", glt->name);
	}

	if (glt->shirt > -1 && glt->pants > -1)
	{
		// create new translation table
		for (i = 0; i < 256; i++)
			translation[i] = i;

		shirt = glt->shirt * 16;

		if (shirt < 128)
			for (i = 0; i < 16; i++)
				translation[TOP_RANGE+i] = shirt + i;
		else
			for (i = 0; i < 16; i++)
				translation[TOP_RANGE+i] = shirt + 15 - i;

		pants = glt->pants * 16;

		if (pants < 128)
			for (i = 0; i < 16; i++)
				translation[BOTTOM_RANGE+i] = pants + i;
		else
			for (i = 0; i < 16; i++)
				translation[BOTTOM_RANGE+i] = pants + 15 - i;

		//translate texture
		size = glt->width * glt->height;
		dst = translated = (byte *) Hunk_Alloc (size);
		src = data;

		for (i = 0; i < size; i++)
			*dst++ = translation[*src++];

		data = translated;
	}

	// fill the source (8-bit only)
	if ((glt->flags & TEXPREF_FLOODFILL) && glt->source_format == SRC_INDEXED)
		Mod_FloodFillSkin (data, glt->source_width, glt->source_height);

	// upload it
	switch (glt->source_format)
	{
	case SRC_INDEXED:
		TM_LoadImage8 (glt, data);
		break;
	case SRC_LIGHTMAP:
		TM_LoadLightmap (glt, data);
		break;
	case SRC_RGBA:
		TM_LoadImage32 (glt, (unsigned *) data);
		break;
	}

	Hunk_FreeToLowMark (mark);
	if (f) fclose (f);
}

/*
================
TM_ReloadImages -- reloads all texture images. called only by vid_restart
================
*/
void R_CreateRTTTextures (int width, int height);

void TM_ReloadImages (void)
{
	gltexture_t *glt;

	for (glt = active_gltextures; glt; glt = glt->next)
	{
		glGenTextures (1, (GLuint *) &glt->texnum);
		GL_BindTexture (GL_TEXTURE0, glt);
		TM_ReloadImage (glt, -1, -1);
	}

	// this is quite shitty
	if (free_gltextures) R_CreateRTTTextures (vid.width, vid.height);
}


/*
================
TM_ReloadNobrightImages -- reloads all texture that were loaded with the nobright palette.  called when gl_fullbrights changes
================
*/
void TM_ReloadNobrightImages (void)
{
	gltexture_t *glt;

	for (glt = active_gltextures; glt; glt = glt->next)
		if (glt->flags & TEXPREF_NOBRIGHT)
			TM_ReloadImage (glt, -1, -1);
}

/*
================================================================================

	TEXTURE BINDING / TEXTURE UNIT SWITCHING

================================================================================
*/

// mh - work properly with multitexture
int	r_currenttexture[3] = {-1, -1, -1}; // to avoid unnecessary texture sets

// default tmu if qglActiveTexture has never been called
GLenum r_currenttmu = GL_TEXTURE0;
static int r_lastmode[3] = {GL_NONE, GL_NONE, GL_NONE};
static GLenum r_lasttype[3] = {GL_TEXTURE_2D, GL_TEXTURE_2D, GL_TEXTURE_2D};
static int r_lastscale[3] = {-1, -1, -1};

void GL_ActiveTexture (GLenum tmu)
{
	if (r_currenttmu == tmu) return;

	r_currenttmu = tmu;
	qglActiveTexture (tmu);
}


void GL_TextureType (GLenum tmu, GLenum type)
{
	int tmunum = tmu - GL_TEXTURE0;

	if (type != r_lasttype[tmunum])
	{
		GL_ActiveTexture (tmu);

		// only disable if there is something to enable
		if (r_lasttype[tmunum] != GL_NONE)
			glDisable (r_lasttype[tmunum]);

		glEnable (type);

		// update type
		r_lasttype[tmunum] = type;

		// force a texenv update irrespective of whether or not it's changed
		r_lastmode[tmunum] = GL_INVALID_VALUE;
		r_lastscale[tmunum] = -1;

		// likewise a texture update
		r_currenttexture[tmunum] = -1;
	}
}


void GL_TexEnv (GLenum tmu, GLenum type, GLenum mode)
{
	int scale = 1;
	int tmunum = (int) tmu - (int) GL_TEXTURE0;

	// activate the texture unit
	GL_ActiveTexture (tmu);
	GL_TextureType (tmu, type);

	if (mode != r_lastmode[tmunum])
	{
		// enable texturing if it was previously disabled
		if (r_lastmode[tmunum] == GL_NONE)
			glEnable (GL_TEXTURE_2D);

		if (mode == GL_NONE)
		{
			// call recursively to revert tmu to default state
			GL_TexEnv (tmu, type, GL_REPLACE);
			glDisable (GL_TEXTURE_2D);
			r_lasttype[tmunum] = GL_REPLACE;	// force a ren-enable of whatever the type was by setting this to an invalid type
			scale = 1;
		}
		else if (mode == GL_RGB_SCALE)
		{
			glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
			glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
			glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_TEXTURE);
			glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_PREVIOUS);
			scale = 2;
		}
		else if (mode == GL_INTERPOLATE)
		{
			glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
			glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_INTERPOLATE);
			glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_TEXTURE);
			glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_PRIMARY_COLOR);
			glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE2_RGB, GL_CONSTANT);
			scale = 1;
		}
		else
		{
			glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, mode);
			scale = 1;
		}

		r_lastmode[tmunum] = mode;
	}

	if (scale != r_lastscale[tmunum])
	{
		glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, (float) scale);
		r_lastscale[tmunum] = scale;
	}
}


/*
================
GL_BindTexture -- johnfitz -- heavy revision

mh - handles switching textures between different tmus more logically
================
*/
void GL_BindTexture (GLenum tmu, gltexture_t *texture)
{
	int tmunum = 0;

	tmunum = tmu - GL_TEXTURE0;
	GL_TextureType (tmu, GL_TEXTURE_2D);

	if (!texture)
	{
		r_currenttexture[tmunum] = -1;
	}
	else if (texture->texnum != r_currenttexture[tmunum])
	{
		// activate the correct tmu
		GL_ActiveTexture (tmu);

		r_currenttexture[tmunum] = texture->texnum;
		glBindTexture (GL_TEXTURE_2D, texture->texnum);
		texture->visframe = r_framecount;
	}
}


// convenience; just unbinds from current textures so that the next GL_BindTexture calls will force an update
// use whenever we want to call glBindTexture ourselves
void GL_UnbindTextures (void)
{
	// unbind all textures
	r_currenttexture[0] = -1;
	r_currenttexture[1] = -1;
	r_currenttexture[2] = -1;

	// unbind cached environments
	GL_TexEnv (GL_TEXTURE0, GL_TEXTURE_2D, GL_NONE);
	GL_TexEnv (GL_TEXTURE1, GL_TEXTURE_2D, GL_NONE);
	GL_TexEnv (GL_TEXTURE2, GL_TEXTURE_2D, GL_NONE);

	// back to TMU0
	GL_ActiveTexture (GL_TEXTURE0);
}

